import requests
import json

from typing import Optional, List

from yaaaf.components.agents.tokens_utils import strip_thought_tokens


class BaseClient:
    async def predict(
        self, messages: "Messages", stop_sequences: Optional[List[str]] = None
    ) -> str:
        """
        Predicts the next message based on the input messages and stop sequences.

        :param messages: The input messages.
        :param stop_sequences: Optional list of stop sequences.
        :return: The predicted message.
        """
        pass


class OllamaClient(BaseClient):
    """Client for Ollama API."""

    def __init__(
        self,
        model: str,
        temperature: float = 0.4,
        max_tokens: int = 2048,
        host: str = "http://localhost:11434",
    ):
        self.model = model
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.host = host

    async def predict(
        self, messages: "Messages", stop_sequences: Optional[List[str]] = None
    ) -> str:
        headers = {"Content-Type": "application/json"}
        data = {
            "model": self.model,
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            "messages": messages.model_dump()["utterances"],
            "options": {
                "stop": stop_sequences,
            },
            "stream": False,
        }
        response = requests.post(
            f"{self.host}/api/chat", headers=headers, data=json.dumps(data)
        )
        if response.status_code == 200:
            return strip_thought_tokens(json.loads(response.text)["message"]["content"])
        else:
            raise Exception(f"Error: {response.status_code}, {response.text}")




from yaaaf.components.data_types import PromptTemplate

goal_extractor_prompt = PromptTemplate(
    prompt="""
You are a goal extractor. Your task is to extract the goal from the given message.
The goal is a specific task that the user wants to accomplish.
The goal should be a single sentence that summarizes the user's latest intention.

Please analyze the following message exchange between the user and the assistant:
<messages>
{messages}
</messages>
"""
)



import re


def strip_thought_tokens(answer: str) -> str:
    answer = re.sub(
        r"<think>(.*?)</think>", "", answer, flags=re.DOTALL | re.MULTILINE
    )
    return answer


def get_first_text_between_tags(text: str, opening_tag: str, closing_tag: str) -> str:
    pattern = f"{opening_tag}(.*?){closing_tag}"
    match = re.search(pattern, text, re.DOTALL | re.MULTILINE)
    if match:
        return match.group(1).strip()

    pattern = f"{opening_tag}(.*)"
    match = re.search(pattern, text, re.DOTALL | re.MULTILINE)
    if match:
        return match.group(1).strip()

    return ""


import os
import sys

from yaaaf.client.run import run_frontend
from yaaaf.server.run import run_server
from yaaaf.variables import get_variables


def print_help():
    print("\n")
    print("These are the available commands:")
    print("> yaaaf backend [port]: start the backend server (default port: 4000)")
    print("> yaaaf frontend [port]: start the frontend server (default port: 3000)")
    print()


def add_cwd_to_syspath():
    sys.path.append(os.getcwd())


def print_incipit():
    print()
    print(f"Running YAAAF version {get_variables()['version']}.")
    print()


def process_cli():
    add_cwd_to_syspath()
    print_incipit()

    arguments = sys.argv
    if len(arguments) >= 2:
        command = arguments[1]
        
        # Use default ports or parse provided port
        if len(arguments) >= 3:
            try:
                port = int(arguments[2])
            except ValueError:
                print("Invalid port number. Must be an integer.\n")
                print_help()
                return
        else:
            # Default ports
            port = 4000 if command == "backend" else 3000

        match command:
            case "backend":
                run_server(host="0.0.0.0", port=port)

            case "frontend":
                run_frontend(port=port)

            case _:
                print("Unknown argument.\n")
                print_help()

    else:
        print("Not enough arguments.\n")
        print_help()


def main():
    try:
        process_cli()

    except RuntimeError as e:
        print(e)
        print()
        print("YAAAF ended due to the exception above.")



import mdpd
import re
import pandas as pd
import json

from typing import Optional, List, Dict, Any

from yaaaf.components.agents.artefacts import Artefact, ArtefactStorage
from yaaaf.components.agents.base_agent import BaseAgent
from yaaaf.components.agents.prompts import tool_agent_prompt_template
from yaaaf.components.agents.settings import task_completed_tag
from yaaaf.components.client import BaseClient
from yaaaf.components.data_types import Messages, PromptTemplate
from yaaaf.connectors.mcp_connector import MCPTools


class ToolAgent(BaseAgent):
    _system_prompt: PromptTemplate = tool_agent_prompt_template
    _completing_tags: List[str] = [task_completed_tag]
    _output_tag = "```tools"
    _stop_sequences = [task_completed_tag]
    _max_steps = 5
    _storage = ArtefactStorage()

    def __init__(self, client: BaseClient, tools: List[MCPTools]):
        self._client = client
        self._tools_description = "\n".join(
            [
                f"Tool group index {group_index}:\n{tool_group.get_tools_descriptions()}\n\n"
                for group_index, tool_group in enumerate(tools)
            ]
        )
        self._tools = tools

    async def query(
        self, messages: Messages, notes: Optional[List[str]] = None
    ) -> str:
        messages = messages.add_system_prompt(
            self._system_prompt.complete(tools_descriptions=self._tools_description)
        )
        all_tool_results = []
        all_tool_calls = []

        for _ in range(self._max_steps):
            answer = await self._client.predict(
                messages=messages, stop_sequences=self._stop_sequences
            )
            if self.is_complete(answer) or answer.strip() == "":
                break

            matches = re.findall(
                rf"{self._output_tag}(.+)```",
                answer,
                re.DOTALL | re.MULTILINE,
            )

            if matches:
                df = mdpd.from_md(matches[0])
                tools_and_args_dict: Dict[str, Any] = df.to_dict("list")
                answer = ""

                for group_index, tool_index, arguments in zip(
                    tools_and_args_dict["group_index"],
                    tools_and_args_dict["tool_index"],
                    tools_and_args_dict["arguments"],
                ):
                    try:
                        # Parse tool index to get group and tool indices
                        tool_group = self._tools[int(group_index)]
                        tool = tool_group[int(tool_index)]

                        # Parse arguments JSON
                        args = (
                            json.loads(arguments)
                            if isinstance(arguments, str)
                            else arguments
                        )

                        # Call the tool
                        result = await tool_group.call_tool(tool.name, args)
                        all_tool_results.append(result)
                        all_tool_calls.append(
                            f"Tool {tool_index} ({tool.name}): {args}"
                        )
                        answer += f"Tool index: {tool_index} -> {result}\n"

                    except Exception as e:
                        error_msg = f"Error calling tool {tool_index}: {str(e)}"
                        all_tool_results.append(error_msg)
                        all_tool_calls.append(f"Tool {tool_index}: ERROR")
                        answer += f"Tool index: {tool_index} -> {error_msg}\n"

                current_output = all_tool_results.copy()
                messages = messages.add_user_utterance(
                    f"The answer is {answer}.\n\nThe output of this query is {current_output}.\n\n\n"
                    f"If the user's answer is answered write {self._completing_tags[0]} at the beginning of your answer.\n"
                    f"Otherwise, try to understand from the answer how to modify the tool calls and get better results.\n"
                )
            else:
                messages = messages.add_user_utterance(
                    f"The answer is {answer} but there is no table.\n"
                    f"If the user's answer is answered write {self._completing_tags[0]} at the beginning of your answer.\n"
                    f"Otherwise, try to understand from the answer how to modify the tool calls and get better results.\n"
                )

        df = pd.DataFrame({"tool_calls": all_tool_calls, "results": all_tool_results})
        tool_id: str = str(hash(str(messages))).replace("-", "")
        self._storage.store_artefact(
            tool_id,
            Artefact(
                type=Artefact.Types.TABLE,
                description=str(messages),
                data=df,
                id=tool_id,
            ),
        )
        return f"The result is in this artefact <artefact type='called-tools-table'>{tool_id}</artefact>"

    def get_description(self) -> str:
        return f"""
Tool agent: This agent uses MCP (Model Context Protocol) tools to perform various operations.
Each tool group contains a collection of tools that can be called with specific arguments.
This agent accepts a query in plain English and uses the appropriate tools to gather information or perform actions.
The tools provide the capabilities needed to answer the user's question or complete the requested task.
        """




import logging
import os
import sys
import re
import sklearn
import numpy as np
import pandas as pd

from io import StringIO
from typing import List, Optional

from yaaaf.components.agents.artefact_utils import (
    get_table_and_model_from_artefacts,
    get_artefacts_from_utterance_content, create_prompt_from_artefacts,
)
from yaaaf.components.agents.artefacts import Artefact, ArtefactStorage
from yaaaf.components.agents.base_agent import BaseAgent
from yaaaf.components.agents.hash_utils import create_hash
from yaaaf.components.agents.settings import task_completed_tag
from yaaaf.components.agents.texts import no_artefact_text
from yaaaf.components.agents.tokens_utils import get_first_text_between_tags
from yaaaf.components.client import BaseClient
from yaaaf.components.data_types import Messages
from yaaaf.components.agents.prompts import mle_agent_prompt_template_without_model, mle_agent_prompt_template_with_model

_logger = logging.getLogger(__name__)


class MleAgent(BaseAgent):
    _completing_tags: List[str] = [task_completed_tag]
    _output_tag = "```python"
    _stop_sequences = _completing_tags
    _max_steps = 5
    _storage = ArtefactStorage()

    def __init__(self, client: BaseClient):
        self._client = client

    async def query(
        self, messages: Messages, notes: Optional[List[str]] = None
    ) -> str:
        last_utterance = messages.utterances[-1]
        artefact_list: List[Artefact] = get_artefacts_from_utterance_content(
            last_utterance.content
        )
        if not artefact_list:
            return no_artefact_text

        hash_string: str = create_hash(str(messages))
        model_name: str = hash_string + ".png"
        messages = messages.add_system_prompt(
            create_prompt_from_artefacts(
                artefact_list,
                model_name,
                mle_agent_prompt_template_with_model,
                mle_agent_prompt_template_without_model,
            )
        )
        df, model = get_table_and_model_from_artefacts(artefact_list)
        code = ""
        code_result = "No code found"
        for _ in range(self._max_steps):
            answer = await self._client.predict(
                messages=messages, stop_sequences=self._stop_sequences
            )
            messages.add_assistant_utterance(answer)
            code = get_first_text_between_tags(answer, self._output_tag, "```")
            if code:
                try:
                    old_stdout = sys.stdout
                    redirected_output = sys.stdout = StringIO()
                    global_variables = globals().copy()
                    global_variables.update({"dataframe": df, "sklearn_model": model})
                    exec(code, global_variables)
                    sys.stdout = old_stdout
                    code_result = redirected_output.getvalue()
                    if code_result.strip() == "":
                        code_result = ""
                except Exception as e:
                    print(e)
                    code_result = f"Error while executing the code above.\nThis exception is raised {str(e)}"
                    answer = str(code_result)

            if (
                self.is_complete(answer)
                or answer.strip() == ""
                or code_result.strip() == ""
            ):
                break

            messages.add_assistant_utterance(
                f"The result is: {code_result}. If there are no errors write {self._completing_tags[0]} at the beginning of your answer.\n"
            )

        if os.path.exists(model_name):
            self._storage.store_artefact(
                hash_string,
                Artefact(
                    type=Artefact.Types.MODEL,
                    description=str(messages),
                    code=code,
                    data=df,
                    id=hash_string,
                ),
            )
            os.remove(model_name)
        return (
            f"The result is in this artefact <artefact type='model'>{hash_string}</artefact>"
            f"Additionally, the model returned the output {code_result}.\n"
        )

    def get_description(self) -> str:
        return f"""
Visualization agent: This agent is given the relevant table, quickly trains a small sklearn model and saves it in a joblib file.
You can use linear interpolation, polynomial regression, SVM, logistic regression, decision trees, random forests.
To call this agent write {self.get_opening_tag()} ENGLISH INSTRUCTIONS AND ARTEFACTS THAT DESCRIBE WHAT TO APPLY THE SKLEARN MODEL TO {self.get_closing_tag()}
The arguments within the tags must be: a) instructions about what to look for in the data 2) the artefacts <artefact> ... </artefact> that describe were found by the other agents above (both tables and models).
The information about what to plot will be then used by the agent.
        """




import hashlib
import logging
import os
import re
from io import StringIO

from typing import Optional, List, Callable

import pandas as pd

from yaaaf.components.agents.artefacts import ArtefactStorage, Artefact
from yaaaf.components.agents.base_agent import BaseAgent
from yaaaf.components.agents.hash_utils import create_hash
from yaaaf.components.agents.settings import task_completed_tag
from yaaaf.components.agents.tokens_utils import get_first_text_between_tags
from yaaaf.components.client import BaseClient
from yaaaf.components.data_types import Messages, PromptTemplate, Note
from yaaaf.components.agents.prompts import sql_agent_prompt_template
from yaaaf.components.sources.sqlite_source import SqliteSource

_path = os.path.dirname(os.path.abspath(__file__))
_logger = logging.getLogger(__name__)


class SqlAgent(BaseAgent):
    _system_prompt: PromptTemplate = sql_agent_prompt_template
    _completing_tags: List[str] = [task_completed_tag]
    _output_tag = "```sql"
    _stop_sequences = [task_completed_tag]
    _max_steps = 5
    _storage = ArtefactStorage()

    def __init__(self, client: BaseClient, source: SqliteSource):
        self._schema = source.get_description()
        self._client = client
        self._source = source

    async def query(
        self, messages: Messages, notes: Optional[List[Note]] = None
    ) -> str:
        try:
            messages = messages.add_system_prompt(
                self._system_prompt.complete(schema=self._schema)
            )
        except Exception as e:
            _logger.error(f"SqlAgent: Failed to add system prompt: {e}")
            raise
        current_output: str | pd.DataFrame = "No output"
        sql_query = "No SQL query"
        for _ in range(self._max_steps):
            try:
                answer = await self._client.predict(
                    messages=messages, stop_sequences=self._stop_sequences
                )
            except Exception as e:
                _logger.error(f"SqlAgent: Client prediction failed in query step: {e}")
                raise
            if self.is_complete(answer) or answer.strip() == "":
                break

            try:
                sql_query = get_first_text_between_tags(answer, self._output_tag, "```")
            except Exception as e:
                _logger.error(f"SqlAgent: Failed to extract SQL query from answer: {e}")
                sql_query = None
            if sql_query:
                if notes is not None:
                    try:
                        note = Note(
                            message=f"```SQL\n{sql_query}\n```",
                            artefact_id=None,
                            agent_name=self.get_name()
                        )
                        notes.append(note)
                    except Exception as e:
                        _logger.error(f"SqlAgent: Failed to create or append SQL note: {e}")
                        # Continue execution even if note creation fails
                try:
                    current_output = self._source.get_data(sql_query)
                except Exception as e:
                    _logger.error(f"SqlAgent: Failed to execute SQL query '{sql_query}': {e}")
                    current_output = f"Error executing SQL: {e}"
                messages = messages.add_user_utterance(
                    f"The answer is {answer}.\n\nThe output of this SQL query is {current_output}.\n\n\n"
                    f"If there are no errors write {self._completing_tags[0]} at the beginning of your answer.\n"
                    f"If there are errors correct the SQL query accordingly you will need to write the SQL query leveraging the schema above.\n"
                )
            else:
                messages = messages.add_user_utterance(
                    f"The answer is {answer} but there is no SQL call. Try again. If there are errors correct the SQL query accordingly."
                )

        try:
            df_info_output = StringIO()
            table_id = create_hash(current_output.to_markdown())
            current_output.info(verbose=True, buf=df_info_output)
            self._storage.store_artefact(
                table_id,
                Artefact(
                    type=Artefact.Types.TABLE,
                    data=current_output,
                    description=df_info_output.getvalue(),
                    code=sql_query,
                ),
            )
        except Exception as e:
            _logger.error(f"SqlAgent: Failed to create or store table artefact: {e}")
            # Return a simple message without artefact if storage fails
            return f"Query completed but failed to store result: {e}"
        return f"The result is in this artifact <artefact type='table'>{table_id}</artefact>."

    def get_description(self) -> str:
        return f"""
SQL agent: This agent calls the relevant sql table and outputs the results.
This agent provides an interface to a dataset through SQL queries. It includes table information and column names.
To call this agent write {self.get_opening_tag()} INFORMATION TO RETRIEVE {self.get_closing_tag()}
Do not write an SQL formula. Just write in clear and brief English the information you need to retrieve.
        """




from typing import List, Dict, Any, Optional
from pydantic import BaseModel
from pydantic_ai.mcp import MCPServerSSE


class ToolDescription(BaseModel):
    name: str
    description: str
    inputSchema: Dict[str, Any]


class MCPTools(BaseModel):
    server_description: str
    tools: List[ToolDescription]
    server: MCPServerSSE

    class Config:
        arbitrary_types_allowed = True

    def __getitem__(self, index: int) -> ToolDescription:
        """Allow indexing to get tools by index"""
        return self.tools[index]

    def __len__(self) -> int:
        """Return number of tools"""
        return len(self.tools)

    def get_tools_descriptions(self) -> str:
        """Get a string representation of available tool descriptions"""
        return "\n".join(
            f"{i}) Name: {tool.name}, Description: {tool.description.strip()}, Input schema: {str(tool.inputSchema).strip()}" for i, tool in enumerate(self.tools)
        )

    async def call_tool(self, tool_name: str, arguments: Dict[str, Any]) -> Any:
        """Call a tool by name with given arguments"""
        if not self.server:
            raise RuntimeError("MCP server connection not available")
        
        return await self.server.call_tool(tool_name, arguments)

    async def call_tool_by_index(self, index: int, arguments: Dict[str, Any]) -> Any:
        """Call a tool by index with given arguments"""
        if index < 0 or index >= len(self.tools):
            raise IndexError(f"Tool index {index} out of range")
        
        tool_name = self.tools[index].name
        return await self.call_tool(tool_name, arguments)


class MCPConnector:
    def __init__(self, url: str, description: str):
        self._url = url
        self._server: Optional[MCPServerSSE] = None
        self._connector_description = description

    async def get_tools(self) -> MCPTools:
        """Connect to MCP server and return Tool object with server and tool descriptions"""
        try:
            self._server = MCPServerSSE(url=self._url)
            await self._server.__aenter__()

            # Get available tools
            tools_response = await self._server.list_tools()
            tool_descriptions = []
            
            for tool_info in tools_response:
                tool_desc = ToolDescription(
                    name=tool_info.name,
                    description=tool_info.description or "",
                    inputSchema=tool_info.parameters_json_schema or {}
                )
                tool_descriptions.append(tool_desc)
            
            # Create Tool object with server reference
            mcp_tools = MCPTools(
                server_description=self._connector_description,
                tools=tool_descriptions,
                server=self._server
            )
            
            return mcp_tools
            
        except Exception as e:
            if self._server:
                await self._server.__aexit__(None, None, None)
            raise e

    async def disconnect(self):
        """Disconnect from MCP server"""
        if self._server:
            await self._server.__aexit__(None, None, None)
            self._server = None



from yaaaf.components.client import BaseClient
from yaaaf.components.data_types import Messages, PromptTemplate
from yaaaf.components.extractors.base_extractor import BaseExtractor
from yaaaf.components.extractors.prompts import goal_extractor_prompt


class GoalExtractor(BaseExtractor):
    """
    GoalExtractor is a class that extracts the goal from a given message.
    It uses a specific prompt template to guide the extraction process.
    """

    _goal_extractor_prompt: PromptTemplate = goal_extractor_prompt

    def __init__(self, client: BaseClient):
        super().__init__()
        self._client = client

    async def extract(self, messages: Messages) -> str:
        instructions = Messages().add_system_prompt(
            self._goal_extractor_prompt.complete(messages=str(messages))
        )
        instructions = instructions.add_user_utterance(
            "Write below the goal in a single sentence"
        )
        answer = await self._client.predict(instructions)
        return answer



import numpy as np

from nltk import download
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize
from typing import List, Tuple
from rank_bm25 import BM25, BM25Okapi

download("stopwords")
download("punkt_tab")


class BM25LocalDB:
    def __init__(self):
        self._indices = []
        self._texts = []
        self._bm25 = None
        self._stopwords = stopwords.words("english")

    def add_text_and_index(self, text: str, index: str):
        self._indices.append(index)
        self._texts.append(word_tokenize(text))

    def get_indices_from_text(
        self, text: str, topn: int
    ) -> Tuple[List[str], List[str]]:
        scores = self._bm25.get_scores(word_tokenize(text))
        best_n = np.argsort(-scores)[:topn]
        return [self._indices[i] for i in best_n], scores[:topn]

    def build(self):
        if self._bm25 is None:
            del self._bm25
        self._bm25 = BM25Okapi(self._texts)



import logging
import re
from typing import List, Tuple, Optional

from yaaaf.components.agents.artefact_utils import get_artefacts_from_utterance_content
from yaaaf.components.agents.artefacts import Artefact, ArtefactStorage
from yaaaf.components.agents.base_agent import BaseAgent
from yaaaf.components.agents.settings import task_completed_tag
from yaaaf.components.client import BaseClient
from yaaaf.components.data_types import Messages, Note
from yaaaf.components.agents.prompts import orchestrator_prompt_template
from yaaaf.components.extractors.goal_extractor import GoalExtractor

_logger = logging.getLogger(__name__)


class OrchestratorAgent(BaseAgent):
    _completing_tags: List[str] = [task_completed_tag]
    _agents_map: {str: BaseAgent} = {}
    _stop_sequences = []
    _max_steps = 10
    _storage = ArtefactStorage()

    def __init__(self, client: BaseClient):
        self._client = client
        self._agents_map = {
            key: agent(client) for key, agent in self._agents_map.items()
        }
        self._goal_extractor = GoalExtractor(client)

    async def query(
        self, messages: Messages, notes: Optional[List[Note]] = None
    ) -> str:
        try:
            messages = messages.add_system_prompt(
                self._get_system_prompt(await self._goal_extractor.extract(messages))
            )
        except Exception as e:
            _logger.error(f"OrchestratorAgent: Failed to extract goal or add system prompt: {e}")
            raise
        
        answer: str = ""
        for step_index in range(self._max_steps):
            try:
                answer = await self._client.predict(
                    messages, stop_sequences=self._stop_sequences
                )
            except Exception as e:
                _logger.error(f"OrchestratorAgent: Client prediction failed at step {step_index}: {e}")
                raise
            agent_to_call, instruction = self.map_answer_to_agent(answer)
            # Extract agent name from tags in the answer, fallback to agent_to_call or orchestrator
            extracted_agent_name = Note.extract_agent_name_from_tags(answer)
            agent_name = extracted_agent_name or (agent_to_call.get_name() if agent_to_call else self.get_name())
            
            if notes is not None:
                try:
                    artefacts = get_artefacts_from_utterance_content(answer)
                    note = Note(
                        message=Note.clean_agent_tags(answer),
                        artefact_id=artefacts[0].id if artefacts else None,
                        agent_name=agent_name
                    )
                    notes.append(note)
                except Exception as e:
                    _logger.error(f"OrchestratorAgent: Failed to create or append note at step {step_index}: {e}")

            if self.is_complete(answer) or answer.strip() == "":
                break
            if agent_to_call is not None:
                if notes is not None:
                    messages = messages.add_assistant_utterance(
                        f"Calling {agent_name} with instruction:\n\n{instruction}\n\n"
                    )
                try:
                    answer = await agent_to_call.query(
                        Messages().add_user_utterance(instruction),
                        notes=notes,
                    )
                except Exception as e:
                    _logger.error(f"OrchestratorAgent: Agent {agent_name} query failed at step {step_index}: {e}")
                    answer = f"Error occurred while calling {agent_name}: {e}"
                try:
                    answer = self._add_relevant_information(answer)
                except Exception as e:
                    _logger.error(f"OrchestratorAgent: Failed to add relevant information at step {step_index}: {e}")

                if notes is not None:
                    try:
                        artefacts = get_artefacts_from_utterance_content(answer)
                        # Extract agent name from tags in the answer, fallback to current agent_name
                        extracted_agent_name = Note.extract_agent_name_from_tags(answer)
                        final_agent_name = extracted_agent_name or agent_name
                        
                        note = Note(
                            message=Note.clean_agent_tags(answer),
                            artefact_id=artefacts[0].id if artefacts else None,
                            agent_name=final_agent_name
                        )
                        notes.append(note)
                    except Exception as e:
                        _logger.error(f"OrchestratorAgent: Failed to create or append agent note at step {step_index}: {e}")

                messages = messages.add_user_utterance(
                    f"The answer from the agent is:\n\n{answer}\n\nWhen you are 100% sure about the answer and the task is done, write the tag {self._completing_tags[0]}."
                )
            else:
                messages = messages.add_assistant_utterance(answer)
                messages = messages.add_user_utterance(
                    f"You didn't call any agent. Is the answer finished or did you miss outputting the tags? Reminder: use the relevant html tags to call the agents.\n\n"
                )
        if not self.is_complete(answer) and step_index == self._max_steps - 1:
            answer += "\nThe Orchestrator agent has finished its maximum number of steps. <task-completed/>"
            if notes is not None:
                notes.append(
                    Note(
                        message="The Orchestrator agent has finished its maximum number of steps.",
                        agent_name=self.get_name()
                    )
                )
        return answer

    def subscribe_agent(self, agent: BaseAgent):
        if agent.get_opening_tag() in self._agents_map:
            raise ValueError(
                f"Agent with tag {agent.get_opening_tag()} already exists."
            )
        self._agents_map[agent.get_opening_tag()] = agent
        self._stop_sequences.append(agent.get_closing_tag())

    def map_answer_to_agent(self, answer: str) -> Tuple[BaseAgent | None, str]:
        for tag, agent in self._agents_map.items():
            if tag in answer:
                matches = re.findall(
                    rf"{agent.get_opening_tag()}(.+)", answer, re.DOTALL | re.MULTILINE
                )
                if matches:
                    return agent, matches[0]

        return None, ""

    def get_description(self) -> str:
        return """
Orchestrator agent: This agent orchestrates the agents.
        """


    def _get_system_prompt(self, goal: str) -> str:
        return orchestrator_prompt_template.complete(
            agents_list="\n".join(
                [
                    "* " + agent.get_description().strip() + "\n"
                    for agent in self._agents_map.values()
                ]
            ),
            all_tags_list="\n".join(
                [
                    agent.get_opening_tag().strip() + agent.get_closing_tag().strip()
                    for agent in self._agents_map.values()
                ]
            ),
            goal=goal,
        )

    def _add_relevant_information(self, answer: str) -> str:
        if "<artefact type='image'>" in answer:
            image_artefact: Artefact = get_artefacts_from_utterance_content(
                answer
            )[0]
            answer = (
                    f"<imageoutput>{image_artefact.id}</imageoutput>"
                    + "\n"
                    + answer
            )
        if "<artefact type='paragraphs-table'>" in answer:
            artefact: Artefact = get_artefacts_from_utterance_content(
                answer
            )[0]
            answer = (
                    f"\n\n{artefact.data.to_markdown(index=False)}\n\n"
                    + answer
            )
        if "<artefact type='called-tools-table'>" in answer:
            artefact: Artefact = get_artefacts_from_utterance_content(
                answer
            )[0]
            answer = (
                    f"\n\n{artefact.data.to_markdown(index=False)}\n\n"
                    + answer
            )
        return answer



import re
from typing import Optional
from pydantic import BaseModel


class Note(BaseModel):
    message: str
    artefact_id: Optional[str] = None
    agent_name: Optional[str] = None

    def __repr__(self):
        return f"Note(message={self.message[:50]}..., artefact_id={self.artefact_id}, agent_name={self.agent_name})"

    def __str__(self):
        return self.__repr__()

    def add_artefact_id(self, artefact_id: str) -> "Note":
        self.artefact_id = artefact_id
        return self

    def add_message(self, message: str) -> "Note":
        self.message = message
        return self
    
    @staticmethod
    def extract_agent_name_from_tags(text: str) -> Optional[str]:
        """Extract agent name from agent tags in the text (e.g., <sqlagent>...</sqlagent>)"""
        # Look for agent tags (excluding artefact and completion tags)
        agent_tag_pattern = r'<(\w+agent)>'
        matches = re.findall(agent_tag_pattern, text, re.IGNORECASE)
        return matches[0] if matches else None
    
    @staticmethod
    def clean_agent_tags(text: str) -> str:
        """Remove agent opening and closing tags while preserving artefact and completion tags"""
        # Remove agent tags (ending with 'agent') but keep artefact and completion tags
        # This preserves <artefact>, <task-completed>, etc.
        cleaned_text = re.sub(r'</?(\w*agent)>', '', text, flags=re.IGNORECASE)
        return cleaned_text.strip()
    
    def set_message_cleaned(self, message: str) -> "Note":
        """Set message after cleaning agent tags"""
        self.message = self.clean_agent_tags(message)
        return self


"""
YAAAF - Yet Another Amazing Agentic Framework

A modular framework for building intelligent agentic applications with Python backend 
and Next.js frontend components. The system features an orchestrator pattern with 
specialized agents for different tasks like SQL queries, web search, visualization, 
machine learning, and reflection.
"""

__version__ = "0.0.3"
__author__ = "YAAAF Contributors"
__email__ = "alberto@fractalego.io"
__license__ = "MIT"

# Core components
from yaaaf.components.data_types import Messages, Note, Utterance, PromptTemplate
from yaaaf.components.agents.base_agent import BaseAgent
from yaaaf.components.agents.orchestrator_agent import OrchestratorAgent
from yaaaf.components.orchestrator_builder import OrchestratorBuilder

# Common agents
from yaaaf.components.agents.sql_agent import SqlAgent
from yaaaf.components.agents.visualization_agent import VisualizationAgent
from yaaaf.components.agents.websearch_agent import DuckDuckGoSearchAgent
from yaaaf.components.agents.reflection_agent import ReflectionAgent
from yaaaf.components.agents.rag_agent import RAGAgent

# Client and configuration
from yaaaf.components.client import BaseClient, OllamaClient
from yaaaf.server.config import get_config

__all__ = [
    # Version info
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    
    # Core data types
    "Messages",
    "Note", 
    "Utterance",
    "PromptTemplate",
    
    # Base classes
    "BaseAgent",
    "BaseClient",
    
    # Main orchestrator
    "OrchestratorAgent",
    "OrchestratorBuilder",
    
    # Specialized agents
    "SqlAgent",
    "VisualizationAgent", 
    "DuckDuckGoSearchAgent",
    "ReflectionAgent",
    "RAGAgent",
    
    # Client implementations
    "OllamaClient",
    
    # Configuration
    "get_config",
]


import re
from typing import List, Optional

from yaaaf.components.agents.base_agent import BaseAgent
from yaaaf.components.agents.settings import task_completed_tag
from yaaaf.components.client import BaseClient
from yaaaf.components.data_types import PromptTemplate, Messages
from yaaaf.components.agents.prompts import reflection_agent_prompt_template


class ReflectionAgent(BaseAgent):
    _system_prompt: PromptTemplate = reflection_agent_prompt_template
    _completing_tags: List[str] = [task_completed_tag]
    _output_tag = "```text"
    _stop_sequences = []
    _max_steps = 5

    def __init__(self, client: BaseClient) -> None:
        self._client = client

    async def query(
        self, messages: Messages, notes: Optional[List[str]] = None
    ) -> str:
        messages = messages.add_system_prompt(self._system_prompt.complete())
        current_output: str = "No output"
        for _ in range(self._max_steps):
            answer = await self._client.predict(
                messages=messages, stop_sequences=self._stop_sequences
            )
            if (
                self._output_tag not in answer and self.is_complete(answer)
            ) or answer.strip() == "":
                break

            messages = messages.add_user_utterance(
                f"The answer is:\n\n{answer}\n\nThink if you need to do more otherwise output {self._completing_tags[0]} at the beginning of your answer.\n"
            )
            matches = re.findall(
                rf"{self._output_tag}(.+)```",
                answer,
                re.DOTALL | re.MULTILINE,
            )
            if matches:
                current_output = matches[0]

        return current_output.replace(task_completed_tag, "")

    def get_description(self) -> str:
        return f"""
Self-reflection agent: This agent thinks step by step about the actions to take.
Use it when you need to think about the task.
Inform the agent about the tools at your disposal (SQL and Visualization).
To call this agent write {self.get_opening_tag()} THINGS TO THINK ABOUT {self.get_closing_tag()}
        """




import os

_path = os.path.dirname(os.path.abspath(__file__))


def run_frontend(port: int):
    server_path = os.path.join(_path, "standalone/apps/www", "server.js")
    os.system(f"node {server_path} --port {str(port)}")



import base64
import logging
import os
import sys
import matplotlib
import re

import numpy as np
import sklearn

from io import StringIO
from typing import List, Optional

from yaaaf.components.agents.artefact_utils import (
    get_table_and_model_from_artefacts,
    get_artefacts_from_utterance_content,
    create_prompt_from_artefacts,
)
from yaaaf.components.agents.artefacts import Artefact, ArtefactStorage
from yaaaf.components.agents.base_agent import BaseAgent
from yaaaf.components.agents.hash_utils import create_hash
from yaaaf.components.agents.settings import task_completed_tag
from yaaaf.components.agents.texts import no_artefact_text
from yaaaf.components.agents.tokens_utils import get_first_text_between_tags
from yaaaf.components.client import BaseClient
from yaaaf.components.data_types import Messages, Note
from yaaaf.components.agents.prompts import (
    visualization_agent_prompt_template_without_model,
    visualization_agent_prompt_template_with_model,
)

_logger = logging.getLogger(__name__)
matplotlib.use("Agg")


class VisualizationAgent(BaseAgent):
    _completing_tags: List[str] = [task_completed_tag]
    _output_tag = "```python"
    _stop_sequences = _completing_tags
    _max_steps = 5
    _storage = ArtefactStorage()

    def __init__(self, client: BaseClient):
        self._client = client

    async def query(
        self, messages: Messages, notes: Optional[List[Note]] = None
    ) -> str:
        try:
            last_utterance = messages.utterances[-1]
            artefact_list: List[Artefact] = get_artefacts_from_utterance_content(
                last_utterance.content
            )
        except Exception as e:
            _logger.error(f"VisualizationAgent: Failed to extract artefacts from utterance: {e}")
            raise
        if not artefact_list:
            return no_artefact_text
        try:
            image_id: str = create_hash(str(messages))
            image_name: str = image_id + ".png"
            messages = messages.add_system_prompt(
                create_prompt_from_artefacts(
                    artefact_list,
                    image_name,
                    visualization_agent_prompt_template_with_model,
                    visualization_agent_prompt_template_without_model,
                )
            )
            df, model = get_table_and_model_from_artefacts(artefact_list)
        except Exception as e:
            _logger.error(f"VisualizationAgent: Failed to setup visualization context: {e}")
            raise
        code = ""
        for _ in range(self._max_steps):
            try:
                answer = await self._client.predict(
                    messages=messages, stop_sequences=self._stop_sequences
                )
            except Exception as e:
                _logger.error(f"VisualizationAgent: Client prediction failed in visualization step: {e}")
                raise
            try:
                messages.add_assistant_utterance(answer)
                code = get_first_text_between_tags(answer, self._output_tag, "```")
            except Exception as e:
                _logger.error(f"VisualizationAgent: Failed to extract code from answer: {e}")
                code = None
            code_result = "No code found"
            if code:
                try:
                    old_stdout = sys.stdout
                    redirected_output = sys.stdout = StringIO()
                    global_variables = globals().copy()
                    global_variables.update({"dataframe": df, "sklearn_model": model})
                    exec(code, global_variables)
                    sys.stdout = old_stdout
                    code_result = redirected_output.getvalue()
                    if code_result.strip() == "":
                        code_result = ""
                except Exception as e:
                    _logger.error(f"VisualizationAgent: Code execution failed: {e}")
                    code_result = f"Error while executing the code above.\nThis exception is raised {str(e)}"
                    answer = str(code_result)

            if (
                self.is_complete(answer)
                or answer.strip() == ""
                or code_result.strip() == ""
            ):
                break

            messages.add_assistant_utterance(
                f"The result is: {code_result}. If there are no errors write {self._completing_tags[0]} at the beginning of your answer.\n"
            )

        if not os.path.exists(image_name):
            _logger.warning(f"VisualizationAgent: No image file generated at {image_name}")
            return "No image was generated. Please try again."

        try:
            with open(image_name, "rb") as file:
                base64_image: str = base64.b64encode(file.read()).decode("ascii")
                self._storage.store_artefact(
                    image_id,
                    Artefact(
                        type=Artefact.Types.IMAGE,
                        image=base64_image,
                        description=str(messages),
                        code=code,
                        data=df,
                        id=image_id,
                    ),
                )
                os.remove(image_name)
        except Exception as e:
            _logger.error(f"VisualizationAgent: Failed to process or store image artefact: {e}")
            # Clean up image file if it exists
            try:
                if os.path.exists(image_name):
                    os.remove(image_name)
            except:
                pass
            return f"Visualization completed but failed to store image: {e}"
        
        result = f"The result is in this artefact <artefact type='image'>{image_id}</artefact>"
        
        if notes is not None:
            try:
                note = Note(
                    message=result,
                    artefact_id=image_id,
                    agent_name=self.get_name()
                )
                notes.append(note)
            except Exception as e:
                _logger.error(f"VisualizationAgent: Failed to create or append visualization note: {e}")
                # Continue execution even if note creation fails
        
        return result

    def get_description(self) -> str:
        return f"""
Visualization agent: This agent is given the relevant artefact table and visualizes the results.
To call this agent write {self.get_opening_tag()} ENGLISH INSTRUCTIONS AND ARTEFACTS THAT DESCRIBE WHAT TO PLOT {self.get_closing_tag()}
The arguments within the tags must be: a) instructions about what to look for in the data 2) the artefacts <artefact> ... </artefact> that describe were found by the other agents above (both tables and models).
The information about what to plot will be then used by the agent.
        """




import re
import pandas as pd

from typing import List, Tuple, Optional
from yaaaf.components.agents.artefacts import Artefact, ArtefactStorage
from yaaaf.components.data_types import Utterance, PromptTemplate


def get_artefacts_from_utterance_content(utterance: Utterance | str) -> List[Artefact]:
    if isinstance(utterance, Utterance):
        utterance_content = utterance.content
    else:
        utterance_content = utterance

    artefact_matches = re.findall(
        rf"<artefact.*?>(.+?)</artefact>",
        utterance_content,
        re.MULTILINE | re.DOTALL,
    )
    if not artefact_matches:
        return []

    storage = ArtefactStorage()
    artefacts: List[Artefact] = []
    for match in artefact_matches:
        artefact_id: str = match
        try:
            artefacts.append(storage.retrieve_from_id(artefact_id))
        except ValueError:
            pass

    return artefacts


def get_table_and_model_from_artefacts(
    artefact_list: List[Artefact],
) -> Tuple["pd.DataFrame", "sklearn.base.BaseEstimator"]:
    table_artefacts = [
        item
        for item in artefact_list
        if item.type == Artefact.Types.TABLE or item.type == Artefact.Types.IMAGE
    ]
    models_artefacts = [
        item for item in artefact_list if item.type == Artefact.Types.MODEL
    ]
    return table_artefacts[0].data if table_artefacts else None, models_artefacts[
        0
    ].model if models_artefacts else None


def create_prompt_from_artefacts(
    artefact_list: List[Artefact],
    filename: str,
    prompt_with_model: PromptTemplate | None,
    prompt_without_model: PromptTemplate,
) -> str:
    table_artefacts = [
        item
        for item in artefact_list
        if item.type == Artefact.Types.TABLE or item.type == Artefact.Types.IMAGE
    ]
    models_artefacts = [
        item for item in artefact_list if item.type == Artefact.Types.MODEL
    ]
    if not table_artefacts:
        table_artefacts = [
            Artefact(
                data=pd.DataFrame(),
                description="",
                type=Artefact.Types.TABLE,
            )
        ]

    if not models_artefacts or not prompt_with_model:
        return prompt_without_model.complete(
            data_source_name="dataframe",
            data_source_type=str(type(table_artefacts[0].data)),
            schema=table_artefacts[0].description,
            filename=filename,
        )

    return prompt_with_model.complete(
        data_source_name="dataframe",
        data_source_type=str(type(table_artefacts[0].data)),
        schema=table_artefacts[0].description,
        model_name="sklearn_model",
        sklearn_model=models_artefacts[0].model,
        training_code=models_artefacts[0].code,
        filename=filename,
    )



import logging
import re
from typing import Optional, Dict, List

import pandas as pd
import sklearn

from pydantic import BaseModel  #
from singleton_decorator import singleton

_logger = logging.getLogger(__name__)


class Artefact(BaseModel):
    model: Optional[sklearn.base.BaseEstimator] = None
    data: Optional[pd.DataFrame] = None
    code: Optional[str] = None
    description: Optional[str] = None
    image: Optional[str] = None
    type: Optional[str] = None
    id: Optional[str] = None

    class Types:
        TABLE = "table"
        IMAGE = "image"
        MODEL = "model"

    class Config:
        arbitrary_types_allowed = True
        use_enum_values = True


@singleton
class ArtefactStorage:
    def __init__(self):
        self.hash_to_artefact_dict: Dict[str, Artefact] = {}

    def store_artefact(self, hash_key: str, artefact: Artefact):
        self.hash_to_artefact_dict[hash_key] = artefact

    def retrieve_from_id(self, hash_key: str) -> Optional[Artefact]:
        if hash_key not in self.hash_to_artefact_dict:
            _logger.warning(f"Artefact with hash {hash_key} not found.")
            raise ValueError(f"Artefact with hash {hash_key} not found.")
        return self.hash_to_artefact_dict.get(hash_key)

    def retrieve_first_from_utterance_string(self, utterance: str) -> Optional[Artefact]:
        artefact_matches = re.findall(
            rf"<artefact.*>(.+?)</artefact>", utterance, re.MULTILINE | re.DOTALL
        )
        if not artefact_matches:
            return None

        return self.retrieve_from_id(artefact_matches[0])

    def retrieve_from_utterance_string(self, utterance: str) -> List[Artefact]:
        artefact_matches = re.findall(
            rf"<artefact.*>(.+?)</artefact>", utterance, re.MULTILINE | re.DOTALL
        )

        return [self.retrieve_from_id(artefact_match) for artefact_match in artefact_matches]





import os
from io import StringIO

import pandas as pd

from duckduckgo_search import DDGS
from typing import Optional, List, Dict

from yaaaf.components.agents.artefacts import ArtefactStorage, Artefact
from yaaaf.components.agents.base_agent import BaseAgent
from yaaaf.components.agents.hash_utils import create_hash
from yaaaf.components.agents.prompts import duckduckgo_search_agent_prompt_template
from yaaaf.components.agents.settings import task_completed_tag
from yaaaf.components.agents.tokens_utils import get_first_text_between_tags
from yaaaf.components.client import BaseClient
from yaaaf.components.data_types import Messages, PromptTemplate

_path = os.path.dirname(os.path.abspath(__file__))


class DuckDuckGoSearchAgent(BaseAgent):
    _system_prompt: PromptTemplate = duckduckgo_search_agent_prompt_template
    _completing_tags: List[str] = [task_completed_tag]
    _output_tag = "```text"
    _stop_sequences = [task_completed_tag]
    _max_steps = 5
    _storage = ArtefactStorage()

    def __init__(self, client: BaseClient):
        self._client = client

    async def query(
        self, messages: Messages, notes: Optional[List[str]] = None
    ) -> str:
        messages = messages.add_system_prompt(
            self._system_prompt
        )
        search_query = ""
        current_output: str | pd.DataFrame = "No output"
        for _ in range(self._max_steps):
            answer = await self._client.predict(
                messages=messages, stop_sequences=self._stop_sequences
            )
            if self.is_complete(answer) or answer.strip() == "":
                break

            search_query: str = get_first_text_between_tags(
                answer, self._output_tag, "```"
            )
            query_results: List[Dict[str, str]] = DDGS().text(search_query, max_results=5)
            if query_results:
                current_output = pd.DataFrame(
                    [[result["title"], result["body"], result["href"]] for result in query_results],
                    columns=["Title", "Summary", "URL"]
                )

                messages = messages.add_user_utterance(
                    f"The web search query was {answer}.\n\nThe result of this query is {current_output}.\n\n\n"
                    f"If there are no errors write {self._completing_tags[0]} at the beginning of your answer.\n"
                    f"If there are errors correct the query accordingly.\n"
                )
            else:
                messages = messages.add_user_utterance(
                    f"The query is {answer} but there are no results from the web search. Try again. If there are errors correct the query accordingly."
                )

        if type(current_output) == str:
            return current_output.replace(task_completed_tag, "")

        df_info_output = StringIO()
        web_search_id = create_hash(str(messages))
        current_output.info(verbose=True, buf=df_info_output)
        self._storage.store_artefact(
            web_search_id,
            Artefact(
                type=Artefact.Types.TABLE,
                data=current_output,
                description=df_info_output.getvalue(),
                code=search_query,
            ),
        )
        return f"The result is in this artifact <artefact type='websearch-result'>{web_search_id}</artefact>."

    def get_description(self) -> str:
        return f"""
Web Search agent: This agent calls a web search engine and outputs the results.
This agent provides an interface to web search engine.
To call this agent write {self.get_opening_tag()} INFORMATION TO RETRIEVE {self.get_closing_tag()}
Just write in clear and brief English the information you need to retrieve between these tags. 
        """




import sqlite3
import pandas as pd

from yaaaf.components.sources.base_source import BaseSource


class SqliteSource(BaseSource):
    def __init__(self, name: str, db_path: str):
        super().__init__(name)
        self.db_path = db_path

    def get_data(self, query: str) -> pd.DataFrame:
        conn = sqlite3.connect(self.db_path)
        try:
            return pd.read_sql_query(query, conn)
        except (ValueError, pd.errors.DatabaseError) as e:
            return pd.DataFrame.from_dict(
                {
                    "Errors": [f"Error in executing SQL query: {e}"],
                    "Results": ["There are no results"],
                }
            )

    def get_description(self) -> str:
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        tables = cursor.execute(
            "SELECT name FROM sqlite_master WHERE type='table';"
        ).fetchall()
        markdown_schema: str = ""
        for table_name in tables:
            table_name = table_name[0].replace(" ", "_")
            table = pd.read_sql_query(f"SELECT * FROM {table_name} LIMIT 1", conn)
            table_columns = [item.replace(" ", "_") for item in table.columns.tolist()]
            markdown_schema += f"# Columns for the table called {table_name} are:\n{table_columns}\n\n\n"
        cursor.close()
        conn.close()
        return markdown_schema

    def ingest(
        self, df: pd.DataFrame, table_name: str, if_exists: str = "replace"
    ) -> None:
        conn = sqlite3.connect(self.db_path)
        df.columns = [col.replace(" ", "_") for col in df.columns]
        df.to_sql(table_name, conn, if_exists=if_exists, index=False)
        conn.close()



import hashlib
from typing import List, Dict

from yaaaf.components.retrievers.local_vector_db import BM25LocalDB
from yaaaf.components.sources.base_source import BaseSource


class RAGSource(BaseSource):
    def __init__(self, description: str, source_path: str):
        self._vector_db = BM25LocalDB()
        self._id_to_chunk: Dict[str, str] = {}
        self._description = description
        self.source_path = source_path

    def add_text(self, text: str):
        node_id: str = hashlib.sha256(text.encode("utf-8")).hexdigest()
        self._vector_db.add_text_and_index(text, node_id)
        self._id_to_chunk[node_id] = text

    def get_data(self, query: str, topn: int = 10) -> List[str]:
        text_ids_and_thresholds = self._vector_db.get_indices_from_text(
            query, topn=topn
        )
        to_return: List[str] = []
        for index in text_ids_and_thresholds[0]:
            to_return.append(self._id_to_chunk[index])
        return to_return

    def get_description(self) -> str:
        self._vector_db.build()
        return self._description



from yaaaf.components.data_types import PromptTemplate


orchestrator_prompt_template = PromptTemplate(
    prompt="""
Your role is to orchestrate a set of 3 analytics agents. You call different agents for different tasks.
These calls happen by writing the name of the agent as the tag name.
Information about the task is provided between tags.

You have these agents at your disposal:
{agents_list}
   
These agents only know what you write between tags and have no memory.
Use the agents to get what you want. Do not write the answer yourself.
The only html tags they understand are these ones: {all_tags_list}. Use these tags to call the agents.

The goal to reach is the following:
{goal}

When you think you have reached this goal, write out the final answer and then {task_completed_tag}
    """
)


sql_agent_prompt_template = PromptTemplate(
    prompt="""
Your task is to write an SQL query according the schema below and the user's instructions
<schema>
{schema}
</schema>
    
In the end, you need to output an SQL instruction string that would retrieve information on an sqlite instance
You can think step-by-step on the actions to take.
However the final output needs to be an SQL instruction string.
This output *must* be between the markdown tags ```sql SQL INSTRUCTION STRING ```
Only give one SQL instruction string per answer.
Limit the number of output rows to 20 at most.
    """
)


reflection_agent_prompt_template = PromptTemplate(
    prompt="""
Your task is to think step by step about the actions to take.
Think about the instructions and creat an action plan to follow them. Be concise and clear.
When you are satisfied with the instructions, you need to output the actions plan between the markdown tags ```text ... ```
"""
)


visualization_agent_prompt_template_without_model = PromptTemplate(
    prompt="""
Your task is to create a Python code that visualises a table as give in the instructions.
The code needs to be written in python between the tags ```python ... ```
The goal of the code is generating and image in matplotlib that explains the data.
This image must be saved in a file named {filename}.
This agent is given in the input information into an already-defined global variable called "{data_source_name}".
This dataframe is of type "{data_source_type}".
The schema of this dataframe is:
<dataframe_schema>
{schema}
</dataframe_schema>
Do not create a new dataframe. Use only the one specified above.

Just save the file, don't show() it.
When you are done output the tag {task_completed_tag}.
"""
)


visualization_agent_prompt_template_with_model = PromptTemplate(
    prompt="""
Your task is to create a Python code that visualises a table as give in the instructions.
The code needs to be written in python between the tags ```python ... ```
The goal of the code is generating and image in matplotlib that explains the data.
This image must be saved in a file named {filename}.
This agent is given in the input information into an already-defined global variable called "{data_source_name}".
This dataframe is of type "{data_source_type}".
The schema of this dataframe is:
<dataframe_schema>
{schema}
</dataframe_schema>
Do not create a new dataframe. Use only the one specified above.

Additionally, the model is given a pre-trained sklearn model into an already-defined global variable called "{model_name}".
<sklearn_model>
{sklearn_model}
</sklearn_model>

This model has been trained using the code
<training_code>
{training_code}
</training_code>
Follow your instructions using the dataframe and the sklearn model to extract the relevant information.

Just save the file, don't show() it.
When you are done output the tag {task_completed_tag}.
"""
)


rag_agent_prompt_template = PromptTemplate(
    prompt="""
Your task is to retrieve information from a collection of texts and pages organised in a list of folders. 
The folders indices with their relative descriptions are given below.
<folders>
{folders}
</folders>

Each folder can be queried with a specific query in plain English.
In the end, you need to output a markdown table with the folder_index and the English query to run on for each folder to answer the user's question.
You can think step-by-step on the actions to take.
However the final output needs to be a markdown table.
This output *must* be between the markdown tags ```retrieved ... ```
The markdown table must have the following columns: 
| folder_index | query |
| ----------- | ----------- |
| ....| ..... | 
    """
)


reviewer_agent_prompt_template_with_model = PromptTemplate(
    prompt="""
Your task is to create a python code that extract the information specified in the instructions. 
The code needs to be written in python between the tags ```python ... ```
The goal of this code is to see if some specific piece of information is in the provided dataframe.

This agent is given the input information into an already-defined global variable called "{data_source_name}".
This dataframe is of type "{data_source_type}".
The schema of this dataframe is:
<dataframe_schema>
{schema}
</dataframe_schema>
Do not create a new dataframe. Use only the one specified above.

Additionally, the model is given a pre-trained sklearn model into an already-defined global variable called "{model_name}".
<sklearn_model>
{sklearn_model}
</sklearn_model>

This model has been trained using the code
<training_code>
{training_code}
</training_code>

Follow your instructions using the dataframe and the sklearn model to extract the relevant information.
When you are done output the tag {task_completed_tag}.
    """
)

reviewer_agent_prompt_template_without_model = PromptTemplate(
    prompt="""
Your task is to create a python code that extract the information specified in the instructions. 
The code needs to be written in python between the tags ```python ... ```
The goal of this code is to see if some specific piece of information is in the provided dataframe.

This agent is given the input information into an already-defined global variable called "{data_source_name}".
This dataframe is of type "{data_source_type}".
The schema of this dataframe is:
<dataframe_schema>
{schema}
</dataframe_schema>
Do not create a new dataframe. Use only the one specified above.

Follow your instructions using the dataframe and the sklearn model to extract the relevant information.
When you are done output the tag {task_completed_tag}.
    """
)


mle_agent_prompt_template_without_model = PromptTemplate(
    prompt="""
Your task is to create a Python code that extracts a trend or finds a patern using sklearn.
The code needs to be written in python between the tags ```python ... ```
The goal of the code is using simple machine learning tools to extract the pattern in the initial instructions.
You will use joblibe to save the sklearn model after it has finished training in a file named {filename}.
This agent is given in the input information into an already-defined global variable called "{data_source_name}".
This dataframe is of type "{data_source_type}".
The schema of this dataframe is:
<dataframe_schema>
{schema}
</dataframe_schema>
Do not create a new dataframe. Use only the one specified above.
When you are done output the tag {task_completed_tag}.
"""
)


mle_agent_prompt_template_with_model = PromptTemplate(
    prompt="""
Your task is to create a Python code that extracts a trend or finds a patern using sklearn.
The code needs to be written in python between the tags ```python ... ```
The goal of the code is using simple machine learning tools to extract the pattern in the initial instructions.
You will use joblibe to save the sklearn model after it has finished training in a file named {filename}.
This agent is given in the input information into an already-defined global variable called "{data_source_name}".
This dataframe is of type "{data_source_type}".
The schema of this dataframe is:
<dataframe_schema>
{schema}
</dataframe_schema>

Additionally, the model is given a pre-trained sklearn model into an already-defined global variable called "{model_name}".
<sklearn_model>
{sklearn_model}
</sklearn_model>

This model has been trained using the code
<training_code>
{training_code}
</training_code>

Do not create a new dataframe. Use only the one specified above.
When you are done output the tag {task_completed_tag}.
"""
)


duckduckgo_search_agent_prompt_template = PromptTemplate(
    prompt="""
Your task is to search the web using DuckDuckGo and find the relevant information.
The query needs to be written in python between the tags ```text ... ```
The goal of this query is to find the relevant information in the web.
DO NOT OUTPUT THE ANSWER YOURSELF. DO NOT WRITE CODE TO CALL THE API.
JUST OUTPUT THE QUERY BETWEEN THE TAGS.
    """
)


url_retriever_agent_prompt_template_without_model = PromptTemplate(
    prompt="""
Your task is to analise markdown table of texts and urls and answer a query given as input.

This agent is given in the markdown table below
<table>
{table}
</table>

Answer the user's query using the table above. 
The output must be a markdown table of paragraphs with the columns: paragraph | relevant_url.
Each paragraph can only use one url as source.
This output *must* be between the markdown tags ```table ... ```.
"""
)



tool_agent_prompt_template = PromptTemplate(
    prompt="""
Your task is to use available MCP tools to help answer the user's question.
The tools available to you are:
<tools>
{tools_descriptions}
</tools>

Each tool can be called with specific arguments based on its input schema.
In the end, you need to output a markdown table with the tool_index and the arguments (as JSON) to call each tool.
You can think step-by-step on the actions to take.
However the final output needs to be a markdown table.
This output *must* be between the markdown tags ```tools ... ```

The table must have the following columns in markdown format:
| group_index | tool_index | arguments |
| ----------- | ----------- | ----------- |
| ....| ..... | .... |

The arguments column should contain valid JSON that matches the tool's input schema.
    """
)



import asyncio
import logging
import threading

from typing import List
from pydantic import BaseModel

from yaaaf.components.agents.artefacts import Artefact, ArtefactStorage
from yaaaf.components.data_types import Utterance, Messages, Note
from yaaaf.components.orchestrator_builder import OrchestratorBuilder
from yaaaf.server.accessories import do_compute, get_utterances
from yaaaf.server.config import get_config

_logger = logging.getLogger(__name__)


class CreateStreamArguments(BaseModel):
    stream_id: str
    messages: List[Utterance]


class NewUtteranceArguments(BaseModel):
    stream_id: str


class ArtefactArguments(BaseModel):
    artefact_id: str


class ArtefactOutput(BaseModel):
    data: str
    code: str
    image: str

    @staticmethod
    def create_from_artefact(artefact: Artefact) -> "ArtefactOutput":
        return ArtefactOutput(
            data=artefact.data.to_html(index=False)
            if artefact.data is not None
            else "",
            code=artefact.code if artefact.code is not None else "",
            image=artefact.image if artefact.image is not None else "",
        )


class ImageArguments(BaseModel):
    image_id: str


def create_stream(arguments: CreateStreamArguments):
    try:
        stream_id = arguments.stream_id
        messages = Messages(utterances=arguments.messages)
        orchestrator = OrchestratorBuilder(get_config()).build()
        t = threading.Thread(target=asyncio.run, args=(do_compute(stream_id, messages, orchestrator),))
        t.start()
    except Exception as e:
        _logger.error(f"Routes: Failed to create stream for {arguments.stream_id}: {e}")
        raise


def get_all_utterances(arguments: NewUtteranceArguments) -> List[Note]:
    try:
        return get_utterances(arguments.stream_id)
    except Exception as e:
        _logger.error(f"Routes: Failed to get utterances for {arguments.stream_id}: {e}")
        raise


def get_artifact(arguments: ArtefactArguments) -> ArtefactOutput:
    try:
        artefact_id = arguments.artefact_id
        artefact_storage = ArtefactStorage(artefact_id)
        artefact = artefact_storage.retrieve_from_id(artefact_id)
        return ArtefactOutput.create_from_artefact(artefact)
    except Exception as e:
        _logger.error(f"Routes: Failed to get artifact {arguments.artefact_id}: {e}")
        raise


def get_image(arguments: ImageArguments) -> str:
    try:
        image_id = arguments.image_id
        artefact_storage = ArtefactStorage(image_id)
        try:
            artefact = artefact_storage.retrieve_from_id(image_id)
            return artefact.image
        except ValueError as e:
            _logger.warning(f"Routes: Artefact with id {image_id} not found: {e}")
            return f"WARNING: Artefact with id {image_id} not found."
    except Exception as e:
        _logger.error(f"Routes: Failed to get image {arguments.image_id}: {e}")
        raise


def get_query_suggestions(query: str) -> List[str]:
    try:
        return get_config().query_suggestions
    except Exception as e:
        _logger.error(f"Routes: Failed to get query suggestions: {e}")
        raise




from typing import Optional, List

from yaaaf.components.data_types import Note


class BaseAgent:
    async def query(
        self, messages: "Messages", notes: Optional[List[Note]] = None
    ) -> str:
        pass

    def get_name(self) -> str:
        return self.__class__.__name__.lower()

    def get_description(self) -> str:
        return "This is just a Base agent. All it does is to say 'Unknown agent'."

    def get_opening_tag(self) -> str:
        return f"<{self.get_name()}>"

    def get_closing_tag(self) -> str:
        return f"</{self.get_name()}>"

    def is_complete(self, answer: str) -> bool:
        if any(tag in answer for tag in self._completing_tags):
            return True

        return False



class BaseExtractor:
    """
    Base class for all extractors.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the extractor with any necessary arguments.
        """
        pass

    def extract(self, *args, **kwargs):
        """
        Extract data from the source.
        """
        raise NotImplementedError("Subclasses must implement this method.")



import logging
import mdpd
from typing import List, Optional

import pandas as pd

from yaaaf.components.agents.artefact_utils import get_table_and_model_from_artefacts, \
    get_artefacts_from_utterance_content
from yaaaf.components.agents.artefacts import Artefact, ArtefactStorage
from yaaaf.components.agents.base_agent import BaseAgent
from yaaaf.components.agents.prompts import url_retriever_agent_prompt_template_without_model
from yaaaf.components.agents.settings import task_completed_tag
from yaaaf.components.agents.texts import no_artefact_text
from yaaaf.components.agents.tokens_utils import get_first_text_between_tags
from yaaaf.components.client import BaseClient
from yaaaf.components.data_types import Messages

_logger = logging.getLogger(__name__)


class UrlReviewerAgent(BaseAgent):
    _completing_tags: List[str] = [task_completed_tag]
    _output_tag = "```table"
    _stop_sequences = _completing_tags
    _max_steps = 5
    _storage = ArtefactStorage()

    def __init__(self, client: BaseClient):
        self._client = client

    async def query(
        self, messages: Messages, notes: Optional[List[str]] = None
    ) -> str:
        last_utterance = messages.utterances[-1]
        artefact_list: List[Artefact] = get_artefacts_from_utterance_content(
            last_utterance.content
        )
        if not artefact_list:
            return no_artefact_text

        df, _ = get_table_and_model_from_artefacts(artefact_list)
        messages = messages.add_system_prompt(
            url_retriever_agent_prompt_template_without_model.complete(
                table=df.to_markdown(index=False),
            )
        )
        output_df: pd.DataFrame = pd.DataFrame()
        for _ in range(self._max_steps):
            answer = await self._client.predict(
                messages=messages, stop_sequences=self._stop_sequences
            )
            messages.add_assistant_utterance(answer)
            output = get_first_text_between_tags(answer, self._output_tag, "```")
            if output.strip() != "":
                output_df = mdpd.from_md(output)

            if (
                self.is_complete(answer)
                or answer.strip() == ""
            ):
                break

            messages.add_assistant_utterance(
                f"The result is <result>{output}</result>. If there the initial query {last_utterance} is answered write {self._completing_tags[0]} at the beginning of your answer.\n"
                f"If there are errors try to correct them in the next steps.\n"
            )
        if output_df.empty:
            return "Could not find a result in the artefacts. Please try again with a different query."

        hash_id: str = str(hash(str(messages))).replace("-", "")
        self._storage.store_artefact(
            hash_id,
            Artefact(
                type=Artefact.Types.TABLE,
                description=str(messages),
                data=output_df,
                id=hash_id,
            ),
        )
        return (
            f"The result is in this artefact <artefact type='paragraphs-table'>{hash_id}</artefact>"
        )

    def get_description(self) -> str:
        return f"""
Url Retriever agent: This agent is given the relevant artefact table of web search results and retrieves the relevant information from these results.
To call this agent write {self.get_opening_tag()} ENGLISH QUERY AND ARTEFACTS THAT DESCRIBE WHAT TO RETRIEVE FROM THE WEB SEARCH RESULTS {self.get_closing_tag()}
This agent is called when you need to better look into the content of a url.
The arguments within the tags must be: a) instructions about what to look for in the data 2) the artefacts <artefact> ... </artefact> that describe were found by the other agents above (only websearch results.
Do *not* use images in the arguments of this agent.
        """




from yaaaf.components.agents.orchestrator_agent import OrchestratorAgent
from yaaaf.components.agents.reflection_agent import ReflectionAgent
from yaaaf.components.agents.reviewer_agent import ReviewerAgent
from yaaaf.components.agents.sql_agent import SqlAgent
from yaaaf.components.agents.url_reviewer_agent import UrlReviewerAgent
from yaaaf.components.agents.visualization_agent import VisualizationAgent
from yaaaf.components.agents.websearch_agent import DuckDuckGoSearchAgent
from yaaaf.components.client import OllamaClient
from yaaaf.components.sources.sqlite_source import SqliteSource
from yaaaf.server.config import Settings


class OrchestratorBuilder:
    def __init__(self, config: Settings):
        self.config = config
        self._agents_map = {
            "reflection": ReflectionAgent,
            "visualization": VisualizationAgent,
            "sql": SqlAgent,
            "reviewer": ReviewerAgent,
            "websearch": DuckDuckGoSearchAgent,
            "url_reviewer": UrlReviewerAgent,
        }

    def build(self):
        client = OllamaClient(
            model=self.config.client.model,
            temperature=self.config.client.temperature,
            max_tokens=self.config.client.max_tokens,
        )
        source = None
        if len(self.config.sources) > 0:
            source = self.config.sources[0]
            sqlite_source = SqliteSource(
                name=source.name,
                db_path=source.path,
            )
        orchestrator = OrchestratorAgent(client)
        for agent_name in self.config.agents:
            if agent_name not in self._agents_map:
                raise ValueError(f"Agent '{agent_name}' is not recognized.")
            if agent_name == "sql" and source is not None:
                orchestrator.subscribe_agent(
                    self._agents_map[agent_name](client=client, source=sqlite_source)
                )
            elif agent_name != "sql":
                orchestrator.subscribe_agent(
                    self._agents_map[agent_name](client=client)
                )
        return orchestrator



import os
from typing import List

from pydantic_settings import BaseSettings


class ClientSettings(BaseSettings):
    model: str | None = None
    temperature: float = 0.5
    max_tokens: int = 1024


class SourceSettings(BaseSettings):
    name: str | None = None
    type: str | None = None
    path: str | None = None


class Settings(BaseSettings):
    client: ClientSettings
    sources: List[SourceSettings] = []
    agents: List[str] = []


def _get_simple_config() -> Settings:
    client_settings: ClientSettings = ClientSettings(
        model="qwen2.5:32b",
        temperature=0.7,
        max_tokens=100,
    )
    config: Settings = Settings(
        client=client_settings,
        sources=[],
        agents=["reflection", "visualization", "sql", "reviewer", "websearch", "url_reviewer"],
    )
    return config


def get_config() -> Settings:
    """
    Returns the default configuration for the orchestrator.
    unless an environment variable `YAAF_CONFIG` is set to a different configuration json file.
    If so, Load that configuration file and return it.
    """
    if os.environ.get("YAAF_CONFIG"):
        config_path = os.environ["YAAF_CONFIG"]
        if not os.path.exists(config_path):
            raise FileNotFoundError(f"Configuration file '{config_path}' does not exist.")
        return Settings.model_validate_json(open(config_path).read())

    return _get_simple_config()


import os
import uvicorn

from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware
from yaaaf.server.routes import (
    create_stream,
    get_artifact,
    get_image,
    get_all_utterances, get_query_suggestions,
)
from yaaaf.server.server_settings import server_settings

app = FastAPI()
app.add_middleware(
    CORSMiddleware, allow_origins=["*"], allow_methods=["*"], allow_headers=["*"]
)
app.add_api_route("/create_stream", endpoint=create_stream, methods=["POST"])
app.add_api_route("/get_utterances", endpoint=get_all_utterances, methods=["POST"])
app.add_api_route("/get_artefact", endpoint=get_artifact, methods=["POST"])
app.add_api_route("/get_image", endpoint=get_image, methods=["POST"])
app.add_api_route("/get_query_suggestions", endpoint=get_query_suggestions, methods=["POST"])

def run_server(host: str, port: int):
    os.environ["YAAF_API_PORT"] = str(port)
    uvicorn.run(app, host=host, port=port)


if __name__ == "__main__":
    os.environ["YAAF_CONFIG"] = "default_config.json"
    run_server(host=server_settings.host, port=server_settings.port)



import mdpd
import re
import pandas as pd

from typing import Optional, List, Dict

from yaaaf.components.agents.artefacts import Artefact, ArtefactStorage
from yaaaf.components.agents.base_agent import BaseAgent
from yaaaf.components.agents.settings import task_completed_tag
from yaaaf.components.client import BaseClient
from yaaaf.components.data_types import Messages, PromptTemplate
from yaaaf.components.agents.prompts import rag_agent_prompt_template
from yaaaf.components.sources.rag_source import RAGSource


class RAGAgent(BaseAgent):
    _system_prompt: PromptTemplate = rag_agent_prompt_template
    _completing_tags: List[str] = [task_completed_tag]
    _output_tag = "```retrieved"
    _stop_sequences = [task_completed_tag]
    _max_steps = 5
    _storage = ArtefactStorage()

    def __init__(self, client: BaseClient, sources: List[RAGSource]):
        self._client = client
        self._folders_description = "\n".join(
            [
                f"Folder index: {index} -> {source.get_description()}"
                for index, source in enumerate(sources)
            ]
        )
        self._sources = sources

    async def query(
        self, messages: Messages, notes: Optional[List[str]] = None
    ) -> str:
        messages = messages.add_system_prompt(
            self._system_prompt.complete(folders=self._folders_description)
        )
        all_retrieved_nodes = []
        all_sources = []
        for _ in range(self._max_steps):
            answer = await self._client.predict(
                messages=messages, stop_sequences=self._stop_sequences
            )
            if self.is_complete(answer) or answer.strip() == "":
                break

            matches = re.findall(
                rf"{self._output_tag}(.+)```",
                answer,
                re.DOTALL | re.MULTILINE,
            )

            if matches:
                df = mdpd.from_md(matches[0])
                retrievers_and_queries_dict: Dict[str, str] = df.to_dict("list")
                answer = ""
                for index, query in zip(
                    retrievers_and_queries_dict["folder_index"],
                    retrievers_and_queries_dict["query"],
                ):
                    source = self._sources[int(index)]
                    retrieved_nodes = source.get_data(query)
                    all_retrieved_nodes.extend(retrieved_nodes)
                    all_sources.extend([source.source_path] * len(retrieved_nodes))
                    answer += f"Folder index: {index} -> {retrieved_nodes}\n"

                current_output = all_retrieved_nodes.copy()
                messages = messages.add_user_utterance(
                    f"The answer is {answer}.\n\nThe output of this query is {current_output}.\n\n\n"
                    f"If the user's answer is answered write {self._completing_tags[0]} at the beginning of your answer.\n"
                    f"Otherwise, try to understand from the answer how to modify the query and get better results.\n"
                )
            else:
                messages = messages.add_user_utterance(
                    f"The answer is {answer} but there is no table.\n"
                    f"If the user's answer is answered write {self._completing_tags[0]} at the beginning of your answer.\n"
                    f"Otherwise, try to understand from the answer how to modify the query and get better results.\n"
                )

        df = pd.DataFrame(
            {"retrieved text chunks": all_retrieved_nodes, "source": all_sources}
        )
        rag_id: str = str(hash(str(messages))).replace("-", "")
        self._storage.store_artefact(
            rag_id,
            Artefact(
                type=Artefact.Types.TABLE,
                description=str(messages),
                data=df,
                id=rag_id,
            ),
        )
        return (
            f"The result is in this artefact <artefact type='table'>{rag_id}</artefact>"
        )

    def get_description(self) -> str:
        return f"""
RAG agent: This agent queries the RAG sources and retrieves the relevant information.
Each source is a folder containing a list of documents.
This agent accepts a query in plain English and returns the relevant documents from the sources.
These documents provide the information needed to answer the user's question.
        """




from typing import List
from pydantic import BaseModel, Field

from yaaaf.components.agents.settings import task_completed_tag


class Utterance(BaseModel):
    role: str = Field(
        ...,
        description="The role of the utterance, e.g., 'user', 'assistant', 'system'",
    )
    content: str = Field(..., description="The content of the utterance")


class PromptTemplate(BaseModel):
    prompt: str = Field(..., description="The prompt template string")

    def complete(self, **kwargs) -> str:
        return self.prompt.replace("{task_completed_tag}", task_completed_tag).format(
            **kwargs
        )


class Messages(BaseModel):
    utterances: List[Utterance] = Field(
        default_factory=list, description="List of utterances in the conversation"
    )

    def add_system_prompt(self, prompt: str | PromptTemplate) -> "Messages":
        if isinstance(prompt, PromptTemplate):
            prompt = prompt.complete()
        system_prompt = Utterance(role="system", content=prompt)
        return Messages(utterances=[system_prompt] + self.utterances)

    def add_assistant_utterance(self, content: str) -> "Messages":
        assistant_utterance = Utterance(role="assistant", content=content)
        return Messages(utterances=self.utterances + [assistant_utterance])

    def add_user_utterance(self, content: str) -> "Messages":
        user_utterance = Utterance(role="user", content=content)
        return Messages(utterances=self.utterances + [user_utterance])

    def __repr__(self):
        return "\n".join(
            [f"{utterance.role}: {utterance.content}" for utterance in self.utterances]
        )

    def __str__(self):
        return self.__repr__()


class BaseSource:
    """
    Base class for all sources.
    """

    def __init__(self, name: str):
        self.name = name

    def get_data(self, query: str) -> str:
        """
        Get data from the source.
        """
        raise NotImplementedError("Subclasses should implement this method.")

    def get_description(self) -> str:
        """
        Get description of the source.
        """
        raise NotImplementedError("Subclasses should implement this method.")



import logging
import os
from typing import Dict, List
from yaaaf.components.agents.orchestrator_agent import OrchestratorAgent
from yaaaf.components.data_types import Note

_path = os.path.dirname(os.path.realpath(__file__))
_logger = logging.getLogger(__name__)
_stream_id_to_messages: Dict[str, List[Note]] = {}


async def do_compute(stream_id, messages, orchestrator: OrchestratorAgent):
    try:
        notes: List[Note] = []
        _stream_id_to_messages[stream_id] = notes
        await orchestrator.query(messages=messages, notes=notes)
    except Exception as e:
        _logger.error(f"Accessories: Failed to compute for stream {stream_id}: {e}")
        # Store error message in notes for frontend
        error_note = Note(
            message=f"Error during computation: {e}",
            artefact_id=None,
            agent_name="system"
        )
        if stream_id in _stream_id_to_messages:
            _stream_id_to_messages[stream_id].append(error_note)
        raise


def get_utterances(stream_id):
    try:
        return _stream_id_to_messages[stream_id]
    except KeyError as e:
        _logger.error(f"Accessories: Stream ID {stream_id} not found in messages: {e}")
        return []
    except Exception as e:
        _logger.error(f"Accessories: Failed to get utterances for stream {stream_id}: {e}")
        raise